HatRack = {}

-- localized strings required to support engineering bags
HatRack.BAG = "Bag" -- 7th return of GetItemInfo on a normal bag
HatRack.ENGINEERING_BAG = "Engineering Bag" -- 7th return of GetItemInfo on an engineering bag
HatRack.REQUIRES_ENGINEERING = "Requires Engineering" -- from tooltip when GetItemInfo ambiguous

function HatRack.LoadDefaults()

	HatRackOptions = HatRackOptions or {
		IconPos = -100,			-- angle of initial minimap icon position
		ShowIcon = "ON",		-- whether to show the minimap button
		SquareMinimap = "OFF",		-- whether the minimap is square instead of circular
		CooldownCount = "OFF",		-- whether to display numerical cooldown counters
		LargeCooldown = "ON",		-- whether cooldown numbers are large or small
		TooltipFollow = "OFF",		-- whether tooltips follow the mouse
		KeepOpen = "OFF",		-- whether menu hides after use
		KeepDocked = "ON",		-- whether to keep menu docked at all times
		Notify = "OFF",			-- whether a message appears when a hat is ready
		DisableToggle="OFF",		-- whether minimap button toggles hat frame
		NotifyUsedOnly="OFF",		-- whether notify happens only on hats used
		NotifyChatAlso="OFF",		-- whether to send notify to chat also
		Locked = "OFF",			-- whether windows can be moved/scaled/rotated
		ShowTooltips = "ON",		-- whether to display tooltips at all
		NotifyThirty = "OFF",		-- whether to notify cooldowns at 30 seconds instead of 0
		MenuOnShift = "OFF",		-- whether menu requires Shift to display
		TinyTooltips = "OFF",		-- whether tooltips display only name and cooldown
		SetColumns = "OFF",		-- whether number of columns in menu is chosen automatically
		Columns = 4,			-- if SetColumns "ON", number of columns before menu wraps
		ShowHotKeys = "OFF",		-- whether hotkeys show on hats
		RedRange = "OFF",		-- whether to monitor and red out out of range hats
		MenuOnRight = "OFF",		-- whether to open menu with right-click
	}

	HatRackPerOptions = HatRackPerOptions or {
		MainDock = "BOTTOMRIGHT",	-- corner of main window docked to
		MenuDock = "BOTTOMLEFT",	-- corner menu window is docked from
		MenuOrient = "VERTICAL",	-- direction of menu window
		MainScale = 1,			-- scaling of main window
		MenuScale = 1,			-- scaling of menu window
		Visible="ON",			-- whether to display the hats
		FirstUse = true,		-- whether this is the first time this user has used the mod
		ItemsUsed = {},			-- table of hats used and their cooldown status
		Alpha = 1,			-- alpha of both windows
		Hidden = {},			-- table of hats hidden
	}
end

--[[ Misc Variables ]]--

HatRack_Version = GetAddOnMetadata("HatRack", "Version") 
BINDING_HEADER_HATRACK = "HatRack"
setglobal("BINDING_NAME_CLICK HatRack_Hat:LeftButton","Use hat")

HatRack.MaxHats = 30 -- add more to HatRack_MenuFrame if this changes
HatRack.BaggedHats = {} -- indexed by number, 1-30 of hats in the menu
HatRack.NumberOfHats = 0 -- number of hats in the menu
HatRack.Corners = {"TOPLEFT", "TOPRIGHT", "BOTTOMLEFT", "BOTTOMRIGHT"}

--[[ Local functions ]]--

-- dock-dependant offset and directions: MainDock..MenuDock
-- x/yoff   = offset MenuFrame is positioned to MainFrame
-- x/ydir   = direction hats are added to menu
-- x/ystart = starting offset when building a menu, relativePoint MenuDock
HatRack.DockStats = {	["TOPRIGHTTOPLEFT"] =		{ xoff=-4, yoff=0,  xdir=1,  ydir=-1, xstart=8,   ystart=-8 },
			["BOTTOMRIGHTBOTTOMLEFT"] =	{ xoff=-4, yoff=0,  xdir=1,  ydir=1,  xstart=8,   ystart=44 },
			["TOPLEFTTOPRIGHT"] =		{ xoff=4,  yoff=0,  xdir=-1, ydir=-1, xstart=-44, ystart=-8 },
			["BOTTOMLEFTBOTTOMRIGHT"] =	{ xoff=4,  yoff=0,  xdir=-1, ydir=1,  xstart=-44, ystart=44 },
			["TOPRIGHTBOTTOMRIGHT"] =	{ xoff=0,  yoff=-4, xdir=-1, ydir=1,  xstart=-44, ystart=44 },
			["BOTTOMRIGHTTOPRIGHT"] =	{ xoff=0,  yoff=4,  xdir=-1, ydir=-1, xstart=-44, ystart=-8 },
			["TOPLEFTBOTTOMLEFT"] =		{ xoff=0,  yoff=-4, xdir=1,  ydir=1,  xstart=8,   ystart=44 },
			["BOTTOMLEFTTOPLEFT"] =		{ xoff=0,  yoff=4,  xdir=1,  ydir=-1, xstart=8,   ystart=-8 } }

-- returns offset and direction depending on current docking. ie: HatRack.DockInfo("xoff")
function HatRack.DockInfo(arg1)
	local anchor = HatRackPerOptions.MainDock..HatRackPerOptions.MenuDock
	if HatRack.DockStats[anchor] and arg1 and HatRack.DockStats[anchor][arg1] then
		return HatRack.DockStats[anchor][arg1]
	else
		return 0
	end
end

-- hide the docking markers
function HatRack.ClearDocking()
	for i=1,4 do
		getglobal("HatRack_MainDock_"..HatRack.Corners[i]):Hide()
		getglobal("HatRack_MenuDock_"..HatRack.Corners[i]):Hide()
	end
end

-- returns true if the two values are close to each other
function HatRack.Near(arg1,arg2)
	return (math.max(arg1,arg2)-math.min(arg1,arg2))<15
end

-- moves the MenuFrame to the dock position against MainFrame
function HatRack.DockWindows()
	HatRack.ClearDocking()
	if HatRackOptions.KeepDocked=="ON" then
		HatRack_MenuFrame:ClearAllPoints()
		if HatRackOptions.Locked=="OFF" then
			HatRack_MenuFrame:SetPoint(HatRackPerOptions.MenuDock,"HatRack_MainFrame",HatRackPerOptions.MainDock,HatRack.DockInfo("xoff"),HatRack.DockInfo("yoff"))
		else
			HatRack_MenuFrame:SetPoint(HatRackPerOptions.MenuDock,"HatRack_MainFrame",HatRackPerOptions.MainDock,HatRack.DockInfo("xoff")*3,HatRack.DockInfo("yoff")*3)
		end
	end
	if HatRack_MenuFrame:IsVisible() then
		HatRack.BuildMenu()
	end
end

-- scan inventory and build MenuFrame
function HatRack.BuildMenu()

	if not IsShiftKeyDown() and HatRackOptions.MenuOnShift=="ON" then
		return
	end

	local idx,i,j,k,texture = 1
	local itemLink,itemID,itemName,equipSlot,itemTexture

	-- go through bags and gather hats into .BaggedHats
	for i=0,4 do
		for j=1,GetContainerNumSlots(i) do
			itemLink = GetContainerItemLink(i,j)
			
			if itemLink then
				_,_,itemID,itemName = string.find(GetContainerItemLink(i,j) or "","item:(%d+).+%[(.+)%]")
				_,_,_,_,_,_,_,_,equipSlot,itemTexture = GetItemInfo(itemID or "")
				if equipSlot=="INVTYPE_HEAD" and (IsAltKeyDown() or not HatRackPerOptions.Hidden[itemID]) then
					if not HatRack.BaggedHats[idx] then
						HatRack.BaggedHats[idx] = {}
					end
					HatRack.BaggedHats[idx].id = itemID
					HatRack.BaggedHats[idx].bag = i
					HatRack.BaggedHats[idx].slot = j
					HatRack.BaggedHats[idx].name = itemName
					HatRack.BaggedHats[idx].texture = itemTexture
					idx = idx + 1
				end
			end
		end
	end
	HatRack.NumberOfHats = math.min(idx-1,HatRack.MaxHats)

	if HatRack.NumberOfHats<1 then
		-- user has no bagged hats :(
		HatRack_MenuFrame:Hide()
	else
		-- display hats outward from docking point
		local col,row,xpos,ypos = 0,0,HatRack.DockInfo("xstart"),HatRack.DockInfo("ystart")
		local max_cols = 1

		if HatRack.NumberOfHats>24 then
			max_cols = 5
		elseif HatRack.NumberOfHats>18 then
			max_cols = 4
		elseif HatRack.NumberOfHats>12 then
			max_cols = 3
		elseif HatRack.NumberOfHats>4 then
			max_cols = 2
		end
		if HatRackOptions.SetColumns=="ON" and HatRackOptions.Columns then
			max_cols = HatRackOptions.Columns
		end

		local item,icon
		for i=1,HatRack.NumberOfHats do
			item = getglobal("HatRack_Menu"..i)
			icon = getglobal("HatRack_Menu"..i.."Icon")
			icon:SetTexture(HatRack.BaggedHats[i].texture)
			if HatRackPerOptions.Hidden[HatRack.BaggedHats[i].id] then
				icon:SetDesaturated(1)
			else
				icon:SetDesaturated(0)
			end
			item:SetPoint("TOPLEFT","HatRack_MenuFrame",HatRackPerOptions.MenuDock,xpos,ypos)

			if HatRackPerOptions.MenuOrient=="VERTICAL" then
				xpos = xpos + HatRack.DockInfo("xdir")*40
				col = col + 1
				if col==max_cols then
					xpos = HatRack.DockInfo("xstart")
					col = 0
					ypos = ypos + HatRack.DockInfo("ydir")*40
					row = row + 1
				end
				item:Show()
			else
				ypos = ypos + HatRack.DockInfo("ydir")*40
				col = col + 1
				if col==max_cols then
					ypos = HatRack.DockInfo("ystart")
					col = 0
					xpos = xpos + HatRack.DockInfo("xdir")*40
					row = row + 1
				end
				item:Show()
			end
		end
		for i=(HatRack.NumberOfHats+1),HatRack.MaxHats do
			getglobal("HatRack_Menu"..i):Hide()
		end
		if col==0 then
			row = row-1
		end

		if HatRackPerOptions.MenuOrient=="VERTICAL" then
			HatRack_MenuFrame:SetWidth(12+(max_cols*40))
			HatRack_MenuFrame:SetHeight(12+((row+1)*40))
		else
			HatRack_MenuFrame:SetWidth(12+((row+1)*40))
			HatRack_MenuFrame:SetHeight(12+(max_cols*40))
		end
		HatRack.UpdateMenuCooldowns()
		HatRack_MenuFrame:Show()
		HatRack.StartTimer("MenuMouseover")
	end

end

function HatRack.Initialize()

	local options = HatRackOptions
	options.KeepDocked = options.KeepDocked or "ON" -- new option for 2.1
	options.Notify = options.Notify or "OFF" -- 2.1
	options.DisableToggle = options.DisableToggle or "OFF" -- new option for 2.2
	options.NotifyUsedOnly = options.NotifyUsedOnly or "OFF" -- 2.2
	options.NotifyChatAlso = options.NotifyChatAlso or "OFF" -- 2.2
	options.ShowTooltips = options.ShowTooltips or "ON" -- 2.3
	options.NotifyThirty = options.NotifyThirty or "OFF" -- 2.5
	options.SquareMinimap = options.SquareMinimap or "OFF" -- 2.6
	options.MenuOnShift = options.MenuOnShift or "OFF" -- 2.6
	options.TinyTooltips = options.TinyTooltips or "OFF" -- 3.0
	options.SetColumns = options.SetColumns or "OFF" -- 3.0
	options.Columns = options.Columns or 4 -- 3.0
	options.LargeCooldown = options.LargeCooldown or "OFF" -- 3.0
	options.ShowHotKeys = options.ShowHotKeys or "OFF" -- 3.0
	HatRackPerOptions.ItemsUsed = HatRackPerOptions.ItemsUsed or {} -- 3.0
	options.HideOnLoad = options.HideOnLoad or "OFF" -- 3.4
	options.RedRange = options.RedRange or "OFF" -- 3.54
	HatRackPerOptions.Alpha = HatRackPerOptions.Alpha or 1 -- 3.5
	HatRackPerOptions.Hidden = HatRackPerOptions.Hidden or {}
	options.MenuOnRight = options.MenuOnRight or "OFF" -- 3.61

	if HatRackPerOptions.MainScale then
		HatRack_MainFrame:SetScale(HatRackPerOptions.MainScale)
	end
	if HatRackPerOptions.MenuScale then
		HatRack_MenuFrame:SetScale(HatRackPerOptions.MenuScale)
	end
	HatRack.ReflectAlpha()

	HatRack_Hat:SetAttribute("type","item")
	HatRack_Hat:SetAttribute("slot",1)
	HatRack_Hat:SetAttribute("shift-slot*",ATTRIBUTE_NOOP)
	HatRack.ReflectMenuOnRight()

	HatRack.InitTimers()
	HatRack.CreateTimer("UpdateWornHats",HatRack.UpdateWornHats,.75)
	HatRack.CreateTimer("DockingMenu",HatRack.DockingMenu,.2,1)
	HatRack.CreateTimer("MenuMouseover",HatRack.MenuMouseover,.25,1)
	HatRack.CreateTimer("TooltipUpdate",HatRack.TooltipUpdate,1,1)
	HatRack.CreateTimer("CooldownUpdate",HatRack.CooldownUpdate,1,1)
	HatRack.CreateTimer("RedRange",HatRack.RedRangeUpdate,.33,1)

	hooksecurefunc("UseInventoryItem", HatRack.newUseInventoryItem)
	hooksecurefunc("UseAction", HatRack.newUseAction)

	HatRack.InitOptions()

	HatRack.UpdateWornHats()
	HatRack.DockWindows()
	HatRack.StartTimer("CooldownUpdate")
	HatRack.ReflectRedRange()

	if HatRackPerOptions.Visible=="ON" and (GetInventoryItemLink("player",1)) then
		HatRack_MainFrame:Show()
	end

	-- fix for OmniCC by N00bZXI
	HatRack_MainFrame:SetFrameLevel(1)
	HatRack_MenuFrame:SetFrameLevel(1)
	HatRack_Hat:SetFrameLevel(2)
	for i=1,30 do
		getglobal("HatRack_Menu"..i):SetFrameLevel(2)
	end

end

function HatRack.ReflectMenuOnRight()
	HatRack_Hat:SetAttribute("slot2",HatRackOptions.MenuOnRight=="ON" and ATTRIBUTE_NOOP or nil)
end

-- returns true if the player is really dead or ghost, not merely FD
function HatRack.IsPlayerReallyDead()
	local dead = UnitIsDeadOrGhost("player")
	if select(2,UnitClass("player"))=="HUNTER" then
		if UnitAura("player","Feign Death") then
			return nil
		end
		for i=1,24 do
			if select(3,UnitAura("player",i))=="Interface\\Icons\\Ability_Rogue_FeignDeath" then
				return nil
			end
		end
	end
	return dead
end

function HatRack.ItemInfo()
	local link,id,name,equipLoc,texture = GetInventoryItemLink("player",1)
	if link then
		local _,_,id = string.find(link,"item:(%d+)")
		name,_,_,_,_,_,_,_,equipLoc,texture = GetItemInfo(id)
	else
		_,texture = GetInventorySlotInfo("HeadSlot")
	end
	return texture,name,equipLoc
end

function HatRack.FindItem(name,includeInventory)
	if includeInventory then
		if GetItemInfo(GetInventoryItemLink("player",1) or "") == name then
			return 1
		end
	end
	for i=0,4 do
		for j=1,GetContainerNumSlots(i) do
			if GetItemInfo(GetContainerItemLink(i,j) or "") == name then
				return nil,i,j
			end
		end
	end
end

--[[ Frame Scripts ]]--

function HatRack.OnLoad(self)

	SlashCmdList["HATRACKCOMMAND"] = HatRack.SlashHandler
	SLASH_HATRACKCOMMAND1 = "/hatrack";
	SLASH_HATRACKCOMMAND2 = "/hat";
	
	self:RegisterEvent("PLAYER_LOGIN")
end

function HatRack.OnEvent(self, event, ...)

	if event=="UNIT_INVENTORY_CHANGED" and select(1,...)=="player" then
		HatRack.UpdateWornHats()
	elseif event=="ACTIONBAR_UPDATE_COOLDOWN" then
		HatRack.UpdateWornCooldowns(1)
	elseif (event=="PLAYER_REGEN_ENABLED" or event=="PLAYER_UNGHOST" or event=="PLAYER_ALIVE") and not HatRack.IsPlayerReallyDead() then
		-- hats can now be swapped after combat/death
		if HatRack.CombatQueue then
			HatRack.EquipHatByName(HatRack.CombatQueue)
			HatRack.CombatQueue = nil
			HatRack.UpdateCombatQueue()
		end
		HatRack_OptMenuOnRight:Enable()
	elseif event=="UPDATE_BINDINGS" then
		HatRack.KeyBindingsChanged()
	elseif event=="PLAYER_REGEN_DISABLED" then
		HatRack_OptMenuOnRight:Disable()
	elseif event=="PLAYER_LOGIN" then
		HatRack.LoadDefaults()
		HatRack.Initialize()
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
		self:RegisterEvent("PLAYER_UNGHOST")
		self:RegisterEvent("PLAYER_ALIVE")
		self:RegisterEvent("UNIT_INVENTORY_CHANGED")
		self:RegisterEvent("UPDATE_BINDINGS")
		self:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN")
	end
end

function HatRack.UpdateWornHats()
	local texture,name = HatRack.ItemInfo()
	HatRack_HatIcon:SetTexture(texture)
	HatRack_HatIcon:SetDesaturated(0)
	HatRack_Hat:SetChecked(0)
	HatRack.UpdateWornCooldowns()
	if HatRack_MenuFrame:IsVisible() then
		HatRack.BuildMenu()
	end
end

function HatRack.SlashHandler(msg)

	msg = string.lower(msg)

	if not msg or msg=="" then
		HatRack.ToggleFrame(HatRack_MainFrame)
	elseif string.find(msg,"^opt") or string.find(msg,"^config") then
		HatRack.ToggleFrame(HatRack_OptFrame)
	elseif msg=="lock" then
		HatRackOptions.Locked="ON"
		HatRack.DockWindows()
		HatRack.ReflectLock()
	elseif msg=="unlock" then
		HatRackOptions.Locked="OFF"
		HatRack.DockWindows()
		HatRack.ReflectLock()
	elseif msg=="reset" then
		HatRack.ResetSettings()
	elseif string.find(msg,"alpha") then
		local _,_,alpha = string.find(msg,"alpha (.+)")
		alpha = tonumber(alpha)
		if alpha and alpha>0 and alpha<=1.0 then
			HatRackPerOptions.Alpha = alpha
		else
			DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00HatRack alpha:")
			DEFAULT_CHAT_FRAME:AddMessage("hat alpha (number) : set alpha from 0.1 to 1.0")
		end
		HatRack.ReflectAlpha()
	elseif string.find(msg,"scale") then
		local _,_,menuscale = string.find(msg,"scale menu (.+)")
		if tonumber(menuscale) then
			HatRack.FrameToScale = HatRack_MenuFrame
			HatRack.ScaleFrame(menuscale)
		end
		local _,_,mainscale = string.find(msg,"scale main (.+)")
		if tonumber(mainscale) then
			HatRack.FrameToScale = HatRack_MainFrame
			HatRack.ScaleFrame(mainscale)
		end
		if not tonumber(menuscale) and not tonumber(mainscale) then
			DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00HatRack scale:")
			DEFAULT_CHAT_FRAME:AddMessage("/hat scale main (number) : set exact main scale")
			DEFAULT_CHAT_FRAME:AddMessage("/hat scale menu (number) : set exact menu scale")
			DEFAULT_CHAT_FRAME:AddMessage("ie, /hat scale menu 0.85")
			DEFAULT_CHAT_FRAME:AddMessage("Note: You can drag the lower-right corner of either window to scale.  This slash command is for those who want to set an exact scale.")
		end
		HatRack.FrameToScale = nil
		HatRackPerOptions.MainScale = HatRack_MainFrame:GetScale()
		HatRackPerOptions.MenuScale = HatRack_MenuFrame:GetScale()
	else
		DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00HatRack useage:")
		DEFAULT_CHAT_FRAME:AddMessage("/hat or /hatmenu : toggle the window")
		DEFAULT_CHAT_FRAME:AddMessage("/hat reset : reset all settings")
		DEFAULT_CHAT_FRAME:AddMessage("/hat opt : summon options window")
		DEFAULT_CHAT_FRAME:AddMessage("/hat lock|unlock : toggles window lock")
		DEFAULT_CHAT_FRAME:AddMessage("/hat scale main|menu (number) : sets an exact scale")
	end
end

function HatRack.ResetSettings()
	StaticPopupDialogs["HATMENURESET"] = {
		text = "Are you sure you want to reset HatRack to default state and reload the UI?",
		button1 = "Yes", button2 = "No", showAlert=1, timeout = 0, whileDead = 1,
		OnAccept = function() HatRackOptions=nil HatRackPerOptions=nil ReloadUI() end,
	}
	StaticPopup_Show("HATMENURESET")
end

--[[ Window Movement ]]--

function HatRack.MainFrame_OnMouseUp(self)
	if GetMouseButtonClicked()=="LeftButton" then
		self:StopMovingOrSizing()
	end
end

function HatRack.MainFrame_OnMouseDown(self)
	if GetMouseButtonClicked()=="LeftButton" and HatRackOptions.Locked=="OFF" then
		self:StartMoving()
	end
end

--[[ Timers ]]

function HatRack.InitTimers()
	HatRack.TimerPool = {}
	HatRack.Timers = {}
end

function HatRack.CreateTimer(name,func,delay,rep)
	HatRack.TimerPool[name] = { func=func,delay=delay,rep=rep,elapsed=delay }
end

function HatRack.IsTimerActive(name)
	for i,j in ipairs(HatRack.Timers) do
		if j==name then
			return i
		end
	end
	return nil
end

function HatRack.StartTimer(name,delay)
	HatRack.TimerPool[name].elapsed = delay or HatRack.TimerPool[name].delay
	if not HatRack.IsTimerActive(name) then
		table.insert(HatRack.Timers,name)
		HatRack_TimersFrame:Show()
	end
end

function HatRack.StopTimer(name)
	local idx = HatRack.IsTimerActive(name)
	if idx then
		table.remove(HatRack.Timers,idx)
		if #(HatRack.Timers)<1 then
			HatRack_TimersFrame:Hide()
		end
	end
end

function HatRack.TimersFrame_OnUpdate(elapsed)
	local timerPool
	for _,name in ipairs(HatRack.Timers) do
		timerPool = HatRack.TimerPool[name]
		timerPool.elapsed = timerPool.elapsed - elapsed
		if timerPool.elapsed < 0 then
			timerPool.func()
			if timerPool.rep then
				timerPool.elapsed = timerPool.delay
			else
				HatRack.StopTimer(name)
			end
		end
	end
end

function HatRack.TimerDebug()
	local on = "|cFF00FF00On"
	local off = "|cFFFF0000Off"
	DEFAULT_CHAT_FRAME:AddMessage("|cFF44AAFFHatRack_TimersFrame is "..(HatRack_TimersFrame:IsVisible() and on or off))
	for i in pairs(HatRack.TimerPool) do
		DEFAULT_CHAT_FRAME:AddMessage(i.." is "..(HatRack.IsTimerActive(i) and on or off))
	end
end

--[[ OnClicks ]]

function HatRack.MainHat_OnClick(self)
	self:SetChecked(0)
	if GetMouseButtonClicked()=="RightButton" and HatRackOptions.MenuOnRight=="ON" then
		if HatRack_MenuFrame:IsVisible() then
			HatRack_MenuFrame:Hide()
		else
			HatRack.BuildMenu()
		end
	elseif IsShiftKeyDown() then
		if ChatFrameEditBox:IsVisible() then
			ChatFrameEditBox:Insert(GetInventoryItemLink("player",1))
		end
	else
		HatRack.ReflectHatUse()
	end
end

function HatRack.MenuHat_OnClick(self)
	self:SetChecked(0)
	local bag,slot = HatRack.BaggedHats[self:GetID()].bag
	local slot = HatRack.BaggedHats[self:GetID()].slot
	if IsShiftKeyDown() and ChatFrameEditBox:IsVisible() then
		ChatFrameEditBox:Insert(GetContainerItemLink(bag,slot))
	elseif IsAltKeyDown() then
		local _,_,itemID = string.find(GetContainerItemLink(bag,slot) or "","item:(%d+)")
		if HatRackPerOptions.Hidden[itemID] then
			HatRackPerOptions.Hidden[itemID] = nil
		else
			HatRackPerOptions.Hidden[itemID] = 1
		end
		HatRack.BuildMenu()
	else
		HatRack.EquipHatByName(HatRack.BaggedHats[self:GetID()].name)
		if not IsShiftKeyDown() and HatRackOptions.KeepOpen=="OFF" then
			HatRack_MenuFrame:Hide()
		end
	end
end

--[[ Docking ]]

function HatRack.MenuFrame_OnMouseDown()
	if GetMouseButtonClicked()=="LeftButton" and HatRackOptions.Locked=="OFF" then
		HatRack_MenuFrame:StartMoving()

		if HatRackOptions.KeepDocked=="ON" then
			HatRack.StartTimer("DockingMenu")
		end
	end
end

function HatRack.MenuFrame_OnMouseUp()
	if GetMouseButtonClicked()=="LeftButton" then
		HatRack.StopTimer("DockingMenu")
		HatRack_MenuFrame:StopMovingOrSizing()
		if HatRackOptions.KeepDocked=="ON" then
			HatRack.DockWindows()
		end
	elseif HatRackOptions.Locked=="OFF" then
		if HatRackPerOptions.MenuOrient=="VERTICAL" then
			HatRackPerOptions.MenuOrient="HORIZONTAL"
		else
			HatRackPerOptions.MenuOrient="VERTICAL"
		end
		HatRack.BuildMenu()
	end
end

function HatRack.DockingMenu()

	local main = HatRack_MainFrame
	local menu = HatRack_MenuFrame
	local mainscale = HatRack_MainFrame:GetScale()
	local menuscale = HatRack_MenuFrame:GetScale()
	local near = HatRack.Near

	if near(main:GetRight()*mainscale,menu:GetLeft()*menuscale) then
		if near(main:GetTop()*mainscale,menu:GetTop()*menuscale) then
			HatRackPerOptions.MainDock = "TOPRIGHT"
			HatRackPerOptions.MenuDock = "TOPLEFT"
		elseif near(main:GetBottom()*mainscale,menu:GetBottom()*menuscale) then
			HatRackPerOptions.MainDock = "BOTTOMRIGHT"
			HatRackPerOptions.MenuDock = "BOTTOMLEFT"
		end
	elseif near(main:GetLeft()*mainscale,menu:GetRight()*menuscale) then
		if near(main:GetTop()*mainscale,menu:GetTop()*menuscale) then
			HatRackPerOptions.MainDock = "TOPLEFT"
			HatRackPerOptions.MenuDock = "TOPRIGHT"
		elseif near(main:GetBottom()*mainscale,menu:GetBottom()*menuscale) then
			HatRackPerOptions.MainDock = "BOTTOMLEFT"
			HatRackPerOptions.MenuDock = "BOTTOMRIGHT"
		end
	elseif near(main:GetRight()*mainscale,menu:GetRight()*menuscale) then
		if near(main:GetTop()*mainscale,menu:GetBottom()*menuscale) then
			HatRackPerOptions.MainDock = "TOPRIGHT"
			HatRackPerOptions.MenuDock = "BOTTOMRIGHT"
		elseif near(main:GetBottom()*mainscale,menu:GetTop()*menuscale) then
			HatRackPerOptions.MainDock = "BOTTOMRIGHT"
			HatRackPerOptions.MenuDock = "TOPRIGHT"
		end
	elseif near(main:GetLeft()*mainscale,menu:GetLeft()*menuscale) then
		if near(main:GetTop()*mainscale,menu:GetBottom()*menuscale) then
			HatRackPerOptions.MainDock = "TOPLEFT"
			HatRackPerOptions.MenuDock = "BOTTOMLEFT"
		elseif near(main:GetBottom()*mainscale,menu:GetTop()*menuscale) then
			HatRackPerOptions.MainDock = "BOTTOMLEFT"
			HatRackPerOptions.MenuDock = "TOPLEFT"
		end
	end
	HatRack.ClearDocking()
	getglobal("HatRack_MainDock_"..HatRackPerOptions.MainDock):Show()
	getglobal("HatRack_MenuDock_"..HatRackPerOptions.MenuDock):Show()
end

function HatRack.MenuMouseover()
	if (not MouseIsOver(HatRack_MainFrame)) and (not MouseIsOver(HatRack_MenuFrame)) and not IsShiftKeyDown() and (HatRackOptions.KeepOpen=="OFF") then
		HatRack.StopTimer("MenuMouseover")
		HatRack_MenuFrame:Hide()
	end
end

--[[ Cooldowns ]]

function HatRack.UpdateWornCooldowns(maybeGlobal)
	local start,duration,enable = GetInventoryItemCooldown("player",1)
	CooldownFrame_SetTimer(HatRack_HatCooldown,start,duration,enable)
	if not maybeGlobal then
		HatRack.WriteWornCooldowns()
	end
end

function HatRack.UpdateMenuCooldowns()
	local start,duration,enable
	for i=1,HatRack.NumberOfHats do
		start,duration,enable = GetContainerItemCooldown(HatRack.BaggedHats[i].bag,HatRack.BaggedHats[i].slot)
		CooldownFrame_SetTimer(getglobal("HatRack_Menu"..i.."Cooldown"),start,duration,enable)
	end
	HatRack.WriteMenuCooldowns()
end

--[[ Item use ]]

function HatRack.ReflectHatUse()
	HatRack_Hat:SetChecked(1)
	HatRack.StartTimer("UpdateWornHats")
	local _,_,id = string.find(GetInventoryItemLink("player",1) or "","item:(%d+)")
	if id then
		HatRackPerOptions.ItemsUsed[id] = 0 -- 0 is an indeterminate state, cooldown will figure if it's worth watching
	end
end

function HatRack.newUseInventoryItem(slot)
	if slot==1 and not MerchantFrame:IsVisible() then
		HatRack.ReflectHatUse()
	end
end

function HatRack.newUseAction(slot,cursor,self)
	if IsEquippedAction(slot) then
		HatRack_TooltipScan:SetOwner(WorldFrame,"ANCHOR_NONE")
		HatRack_TooltipScan:SetAction(slot)
		local _,hat = HatRack.ItemInfo(1)
		if GameTooltipTextLeft1:GetText()==hat then
			HatRack.ReflectHatUse()
		end
	end
end

--[[ Tooltips ]]

function HatRack.WornHatTooltip(self)
	if HatRackOptions.ShowTooltips=="OFF" then
		return
	end
	HatRack.TooltipOwner = self
	HatRack.TooltipType = "INVENTORY"
	HatRack.TooltipBag = HatRack.CombatQueue
	HatRack.StartTimer("TooltipUpdate",0)
end

function HatRack.MenuHatTooltip(self)
	local id = self:GetID()
	if HatRackOptions.ShowTooltips=="OFF" then
		return
	end
	HatRack.TooltipOwner = self
	HatRack.TooltipType = "BAG"
	HatRack.TooltipBag = HatRack.BaggedHats[id].bag
	HatRack.TooltipSlot = HatRack.BaggedHats[id].slot
	HatRack.StartTimer("TooltipUpdate",0)
end

function HatRack.ClearTooltip()
	GameTooltip:Hide()
	HatRack.StopTimer("TooltipUpdate")
	HatRack.TooltipType = nil
end

function HatRack.AnchorTooltip(self)
	if HatRackOptions.TooltipFollow=="ON" then
		if self.GetLeft and self:GetLeft() and self:GetLeft()<400 then
			GameTooltip:SetOwner(self,"ANCHOR_RIGHT")
		else
			GameTooltip:SetOwner(self,"ANCHOR_LEFT")
		end
	else
		GameTooltip_SetDefaultAnchor(GameTooltip,self)
	end
end

-- updates the tooltip created in the functions above
function HatRack.TooltipUpdate()
	if HatRack.TooltipType then
		local cooldown
		HatRack.AnchorTooltip(HatRack.TooltipOwner)
		if HatRack.TooltipType=="BAG" then
			GameTooltip:SetBagItem(HatRack.TooltipBag,HatRack.TooltipSlot)
			cooldown = GetContainerItemCooldown(HatRack.TooltipBag,HatRack.TooltipSlot)
		else
			GameTooltip:SetInventoryItem("player",1)
			cooldown = GetInventoryItemCooldown("player",1)
		end
		HatRack.ShrinkTooltip(HatRack.TooltipOwner) -- if TinyTooltips on, shrink it
		if HatRack.TooltipType=="INVENTORY" and HatRack.TooltipBag then
			GameTooltip:AddLine("Queued: "..HatRack.TooltipBag)
		end
		GameTooltip:Show()
		if cooldown==0 then
			-- stop updates if this hat has no cooldown
			HatRack.TooltipType = nil
			HatRack.StopTimer("TooltipUpdate")
		end
	end

end

-- normal tooltip for options
function HatRack.OnTooltip(self,line1,line2)
	if HatRackOptions.ShowTooltips=="ON" then
		HatRack.AnchorTooltip(self)
		if line1 then
			GameTooltip:AddLine(line1)
			GameTooltip:AddLine(line2,.8,.8,.8,1)
			GameTooltip:Show()
		else
			local name = self:GetName() or ""
			for i=1,#(HatRack.CheckOptInfo) do
				if name=="HatRack_Opt"..HatRack.CheckOptInfo[i][1] and HatRack.CheckOptInfo[i][3] then
					HatRack.OnTooltip(self,HatRack.CheckOptInfo[i][3],HatRack.CheckOptInfo[i][4])
				end
			end
			for i=1,#(HatRack.TooltipInfo) do
				if HatRack.TooltipInfo[i][1]==name and HatRack.TooltipInfo[i][2] then
					HatRack.OnTooltip(self,HatRack.TooltipInfo[i][2],HatRack.TooltipInfo[i][3])
				end
			end
		end
	end
end

-- strip format reordering in global strings
HatRack.ITEM_SPELL_CHARGES = string.gsub(ITEM_SPELL_CHARGES,"%%%d%$d","%%d")

function HatRack.ShrinkTooltip(owner)
	if HatRackOptions.TinyTooltips=="ON" then
		local r,g,b = GameTooltipTextLeft1:GetTextColor()
		local name = GameTooltipTextLeft1:GetText()
		local line,cooldown,charge
		for i=2,GameTooltip:NumLines() do
			line = getglobal("GameTooltipTextLeft"..i)
			if line:IsVisible() then
				line = line:GetText() or ""
				if string.find(line,COOLDOWN_REMAINING) then
					cooldown = line
				end
				if string.find(line,HatRack.ITEM_SPELL_CHARGES) then
					charge = line
				end
			end
		end
		HatRack.AnchorTooltip(owner)
		GameTooltip:AddLine(name,r,g,b)
		GameTooltip:AddLine(charge,1,1,1)
		GameTooltip:AddLine(cooldown,1,1,1)
	end
end

-- returns 1 if the item at bag(,slot) is an engineered hat
function HatRack.IsEngineered(bag,slot)
	local item = slot and GetContainerItemLink(bag,slot) or GetInventoryItemLink("player",1)
	if item then
		HatRack_TooltipScan:SetOwner(WorldFrame,"ANCHOR_NONE")
		HatRack_TooltipScan:SetHyperlink(item)
		for i=1,HatRack_TooltipScan:NumLines() do
			if string.match(getglobal("HatRack_TooltipScanTextLeft"..i):GetText() or "",HatRack.REQUIRES_ENGINEERING) then
				return 1
			end
		end
	end
end

-- returns bag,slot of a free bag space, if one found.  engineering true if only looking for an engineering bag
function HatRack.FindSpace(engineering)
	local bagType
	for i=4,0,-1 do
		bagType = select(7,GetItemInfo(GetInventoryItemLink("player",19+i) or ""))
		if (engineering and bagType==HatRack.ENGINEERING_BAG) or (not engineering and bagType==HatRack.BAG) then
			for j=1,GetContainerNumSlots(i) do
				if not GetContainerItemLink(i,j) then
					return i,j
				end
			end
		end
	end
end

--[[ Combat Queue ]]

function HatRack.EquipHatByName(name)
	if not name then return end
	if UnitAffectingCombat("player") or HatRack.IsPlayerReallyDead() then
		-- queue hat
		if HatRack.CombatQueue==name and not imperative then
			HatRack.CombatQueue = nil
		else
			HatRack.CombatQueue = name
		end
	elseif not CursorHasItem() and not SpellIsTargeting() then
		local _,b,s = HatRack.FindItem(name)
		if b then
			local _,_,isLocked = GetContainerItemInfo(b,s)
			if not isLocked and not IsInventoryItemLocked(1) then
				-- neither container item nor inventory item locked, perform swap
				local directSwap = 1 -- assume a direct swap will happen
				if select(7,GetItemInfo(GetInventoryItemLink("player",19+b) or ""))==HatRack.ENGINEERING_BAG then
					-- incoming hat is in an engineering bag
					if not HatRack.IsEngineered(1) then
						-- outgoing hat can't go inside it
						local freeBag,freeSlot = HatRack.FindSpace()
						if freeBag then
							PickupInventoryItem(1)
							PickupContainerItem(freeBag,freeSlot)
							PickupContainerItem(b,s)
							EquipCursorItem(1)
							directSwap = nil
						end
					end
				elseif HatRack.IsEngineered(1) and not HatRack.IsEngineered(b,s) then
					-- outgoing hat is engineered, incoming hat is not
					local freeBag,freeSlot = HatRack.FindSpace(1)
					if freeBag then
						-- move outgoing hat to engineering bag, equip incoming hat
						PickupInventoryItem(1)
						PickupContainerItem(freeBag,freeSlot)
						PickupContainerItem(b,s)
						EquipCursorItem(1)
						directSwap = nil
					end
				end
				if directSwap then
					PickupContainerItem(b,s)
					PickupInventoryItem(1)
				end
				HatRack_HatIcon:SetDesaturated(1)
				HatRack.StartTimer("UpdateWornHats") -- in case it's not equipped (stunned, etc)
			end
		end
	end
	HatRack.UpdateCombatQueue()
end	

function HatRack.UpdateCombatQueue()
	local bag,slot
	local hat = HatRack.CombatQueue
	local icon = HatRack_HatQueue
	icon:Hide()
	if hat then
		_,bag,slot = HatRack.FindItem(hat)
		if bag then
			icon:SetTexture(GetContainerItemInfo(bag,slot))
			icon:Show()
		end
	end
end

--[[ Notify ]]

function HatRack.Notify(msg)
	PlaySound("GnomeExploration")
	if SCT_Display then -- send via SCT if it exists
		SCT_Display(msg,{r=.2,g=.7,b=.9})
	elseif SHOW_COMBAT_TEXT=="1" then
		CombatText_AddMessage(msg, CombatText_StandardScroll, .2, .7, .9) -- or default UI's SCT
	else
		-- send vis UIErrorsFrame if neither SCT exists
		UIErrorsFrame:AddMessage(msg,.2,.7,.9,1,UIERRORS_HOLD_TIME)
	end
	if HatRackOptions.NotifyChatAlso=="ON" then
		DEFAULT_CHAT_FRAME:AddMessage("|cff33b2e5"..msg)
	end
end

function HatRack.CooldownUpdate()
	local inv,bag,slot,start,duration,name,remain
	for i in pairs(HatRackPerOptions.ItemsUsed) do
		start,duration = GetItemCooldown(i)
		if start and HatRackPerOptions.ItemsUsed[i]<3 then
			HatRackPerOptions.ItemsUsed[i] = HatRackPerOptions.ItemsUsed[i] + 1 -- count for 3 seconds before seeing if this is a real cooldown
		elseif start then
			if start>0 then
				remain = duration - (GetTime()-start)
				if HatRackPerOptions.ItemsUsed[i]<5 then
					if remain>29 then
						HatRackPerOptions.ItemsUsed[i] = 30 -- first actual cooldown greater than 30 seconds, tag it for 30+0 notify
					elseif remain>5 then
						HatRackPerOptions.ItemsUsed[i] = 5 -- first actual cooldown less than 30 but greater than 5, tag for 0 notify
					end
				end
			end
			if HatRackPerOptions.ItemsUsed[i]==30 and start>0 and remain<30 then
				if HatRackOptions.NotifyThirty=="ON" then
					name = GetItemInfo(i)
					if name then
						HatRack.Notify(name.." ready soon!")
					end
				end
				HatRackPerOptions.ItemsUsed[i]=5 -- tag for just 0 notify now
			elseif HatRackPerOptions.ItemsUsed[i]==5 and start==0 then
				if HatRackOptions.Notify=="ON" then
					name = GetItemInfo(i)
					if name then
						HatRack.Notify(name.." ready!")
					end
				end
			end
			if start==0 then
				HatRackPerOptions.ItemsUsed[i] = nil
			end
		end
	end

	-- update cooldown numbers
	if HatRackOptions.CooldownCount=="ON" then
		if HatRack_MainFrame:IsVisible() then
			HatRack.WriteWornCooldowns()
		end
		if HatRack_MenuFrame:IsVisible() then
			HatRack.WriteMenuCooldowns()
		end
	end
end

function HatRack.WriteWornCooldowns()
	local start, duration
	start, duration = GetInventoryItemCooldown("player",1)
	HatRack.WriteCooldown(HatRack_HatTime,start,duration)
end

function HatRack.WriteMenuCooldowns()
	local start, duration
	for i=1,HatRack.NumberOfHats do
		start, duration = GetContainerItemCooldown(HatRack.BaggedHats[i].bag,HatRack.BaggedHats[i].slot)
		HatRack.WriteCooldown(getglobal("HatRack_Menu"..i.."Time"),start,duration)
	end
end

function HatRack.WriteCooldown(where,start,duration)
	local cooldown = duration - (GetTime()-start)
	if start==0 or HatRackOptions.CooldownCount=="OFF" then
		where:SetText("")
	elseif cooldown<3 and not where:GetText() then
		-- this is a global cooldown. don't display it. not accurate but at least not annoying
	else
		where:SetText((cooldown<60 and math.floor(cooldown+.5).." s") or (cooldown<3600 and math.ceil(cooldown/60).." m") or math.ceil(cooldown/3600).." h")
	end
end

function HatRack.OnShow()
	HatRackPerOptions.Visible = "ON"
	if HatRackOptions.KeepOpen=="ON" then
		HatRack.BuildMenu()
	end
end

function HatRack.OnHide()
	HatRackPerOptions.Visible = "OFF"
	HatRack_MenuFrame:Hide()
end

function HatRack.ReflectAlpha()
	HatRack_MainFrame:SetAlpha(HatRackPerOptions.Alpha)
	HatRack_MenuFrame:SetAlpha(HatRackPerOptions.Alpha)
end

--[[ Key bindings ]]

function HatRack.KeyBindingsChanged()
	if HatRackOptions.ShowHotKeys=="ON" then
		local key
		for i=0,1 do
			local key = GetBindingKey("CLICK HatRack_Hat:LeftButton")
			HatRack_HatHotKey:SetText(GetBindingText(key or "",nil,1))
		end
	else
		HatRack_HatHotKey:SetText("")
	end
end

--[[ Monitor Range ]]

function HatRack.ReflectRedRange()
	if HatRackOptions.RedRange=="ON" then
		HatRack.StartTimer("RedRange")
	else
		HatRack.StopTimer("RedRange")
		HatRack_HatIcon:SetVertexColor(1,1,1)
	end
end

function HatRack.RedRangeUpdate()
	local item = GetInventoryItemLink("player",1)
	if item and IsItemInRange(item)==0 then
		HatRack_HatIcon:SetVertexColor(1,.3,.3)
	else
		HatRack_HatIcon:SetVertexColor(1,1,1)
	end
end
