--[[ HatRackOpt.lua : Options and sort window for HatRack ]]

HatRack.CheckOptInfo = {
	{"ShowIcon","ON","Minimap button","Show or hide minimap button."},
	{"SquareMinimap","OFF","Square minimap","Move minimap button as if around a square minimap.","ShowIcon"},
	{"CooldownCount","OFF","Cooldown numbers","Display time remaining on cooldowns ontop of the button."},
	{"TooltipFollow","OFF","At cursor","Display all tooltips at the cursor.","ShowTooltips"},
	{"KeepOpen","OFF","Keep menu open","Keep menu open at all times."},
	{"KeepDocked","ON","Keep menu docked","Keep menu docked at all times."},
	{"Notify","OFF","Notify when ready","Sends an overhead notification when a hat's cooldown is complete."},
	{"DisableToggle","OFF","Disable toggle","Disables the minimap button's ability to toggle the hat frame.","ShowIcon"},
	{"NotifyChatAlso","OFF","Notify chat also","Sends notifications through chat also."},
	{"Locked","OFF","Lock windows","Prevents the windows from being moved, resized or rotated."},
	{"ShowTooltips","ON","Show tooltips","Shows tooltips."},
	{"NotifyThirty","ON","Notify at 30 sec","Sends an overhead notification when a hat has 30 seconds left on cooldown."},
	{"MenuOnShift","OFF","Menu on shift","Check this to prevent the menu appearing unless Shift is held."},
	{"TinyTooltips","OFF","Tiny tooltips","Shrink hat tooltips to only their name, charges and cooldown.","ShowTooltips"},
	{"SetColumns","OFF","Wrap at","Define how many hats before the menu will wrap to the next row.\n\nUncheck to let HatRack choose how to wrap the menu."},
	{"LargeCooldown","ON","Large numbers","Display the cooldown time in a larger font.","CooldownCount"},
	{"ShowHotKeys","ON","Show key bindings","Display the key bindings over the equipped hats."},
	{"RedRange","OFF","Red out of range","Check this to red out worn hats that are out of range to a valid target.  ie, Goblin Rocket Helmet."},
	{"MenuOnRight","OFF","Menu on right-click","Check this to prevent the menu from appearing until either worn hat is right-clicked.\n\nNOTE: This setting CANNOT be changed while in combat."}
}

-- table.insert(HatRack.CheckOptInfo,)

HatRack.TooltipInfo = {
	{"HatRack_LockButton","Lock windows","Prevents the windows from being moved, resized or rotated."},
}

function HatRack.InitOptions()
	HatRack.CreateTimer("DragMinimapButton",HatRack.DragMinimapButton,0,1)
	HatRack.MoveMinimapButton()
	local item
	for i=1,#(HatRack.CheckOptInfo) do
		item = getglobal("HatRack_Opt"..HatRack.CheckOptInfo[i][1].."Text")
		if item then
			item:SetText(HatRack.CheckOptInfo[i][3])
			item:SetTextColor(.95,.95,.95)
		end
	end
	table.insert(UISpecialFrames,"HatRack_OptFrame")
	HatRack_Title:SetText("HatRack "..HatRack_Version)

	HatRack_OptFrame:SetBackdropBorderColor(.3,.3,.3,1)
	HatRack_OptFrame:SetHeight(306)
	HatRack_SubOptFrame:SetBackdropBorderColor(.3,.3,.3,1)
	HatRack_SubOptFrame:SetPoint("TOPLEFT",HatRack_OptFrame,"TOPLEFT",8,-24)
	HatRack_OptColumnsSlider:SetValue(HatRackOptions.Columns)
	HatRack_OptMainScaleSlider:SetValue(HatRackPerOptions.MainScale)
	HatRack_OptMenuScaleSlider:SetValue(HatRackPerOptions.MenuScale)
	HatRack.ReflectLock()
	HatRack.ReflectCooldownFont()
	HatRack.KeyBindingsChanged()
end

function HatRack.ToggleFrame(frame)
	if frame:IsVisible() then
		frame:Hide()
	else
		frame:Show()
	end
end

function HatRack.OptFrame_OnShow()
	HatRack.ValidateChecks()
end

--[[ Minimap button ]]

function HatRack.MoveMinimapButton()
	local xpos,ypos
	if HatRackOptions.SquareMinimap=="ON" then
		xpos = 110 * cos(HatRackOptions.IconPos or 0)
		ypos = 110 * sin(HatRackOptions.IconPos or 0)
		xpos = math.max(-82,math.min(xpos,84))
		ypos = math.max(-86,math.min(ypos,82))
	else
		xpos = 80*cos(HatRackOptions.IconPos or 0)
		ypos = 80*sin(HatRackOptions.IconPos or 0)
	end
	HatRack_IconFrame:SetPoint("TOPLEFT","Minimap","TOPLEFT",52-xpos,ypos-52)
	if HatRackOptions.ShowIcon=="ON" then
		HatRack_IconFrame:Show()
	else
		HatRack_IconFrame:Hide()
	end
end

function HatRack.DragMinimapButton()
	local xpos,ypos = GetCursorPosition()
	local xmin,ymin = Minimap:GetLeft() or 400, Minimap:GetBottom() or 400
	xpos = xmin-xpos/Minimap:GetEffectiveScale()+70
	ypos = ypos/Minimap:GetEffectiveScale()-ymin-70
	HatRackOptions.IconPos = math.deg(math.atan2(ypos,xpos))
	HatRack.MoveMinimapButton()
end

function HatRack.MinimapButton_OnClick()
	PlaySound("GAMEGENERICBUTTONPRESS")
	if IsShiftKeyDown() then
		HatRackOptions.Locked = HatRackOptions.Locked=="ON" and "OFF" or "ON"
		HatRack.ReflectLock()
	else
		if arg1=="LeftButton" and HatRackOptions.DisableToggle=="OFF" then
			HatRack.ToggleFrame(HatRack_MainFrame)
		else
			HatRack.ToggleFrame(HatRack_OptFrame)
		end
	end
end

--[[ CheckButton ]]

function HatRack.ValidateChecks()
	local check,button
	for i=1,#(HatRack.CheckOptInfo) do
		check = HatRack.CheckOptInfo[i]
		button = getglobal("HatRack_Opt"..check[1])
		if button then
			button:SetChecked(HatRackOptions[check[1]]=="ON")
			if check[5] then
				if HatRackOptions[check[5]]=="ON" then
					button:Enable()
					getglobal("HatRack_Opt"..check[1].."Text"):SetTextColor(.95,.95,.95)
				else
					button:Disable()
					getglobal("HatRack_Opt"..check[1].."Text"):SetTextColor(.5,.5,.5)
				end
			end
		end
	end
	HatRack_OptColumnsSlider:SetAlpha((HatRackOptions.SetColumns=="ON") and 1 or .5)
	HatRack_OptColumnsSlider:EnableMouse((HatRackOptions.SetColumns=="ON") and 1 or 0)
	HatRack_OptColumnsSlider:SetValue(HatRackOptions.Columns)
end

function HatRack.OptColumnsSlider_OnValueChanged(self)
	if HatRackOptions then
		HatRackOptions.Columns = self:GetValue()
		HatRack_OptColumnsSliderText:SetText(HatRackOptions.Columns.." hats")
		if HatRack_MenuFrame:IsVisible() then
			HatRack.BuildMenu()
		end
	end
end

function HatRack.OptMainScaleSlider_OnValueChanged(self)
	if HatRackPerOptions then
		HatRackPerOptions.MainScale = self:GetValue()
		HatRack_OptMainScaleSliderText:SetText(format("Main Scale: %.2f",HatRackPerOptions.MainScale))
		HatRack_MainFrame:SetScale(HatRackPerOptions.MainScale)
	end
end

function HatRack.OptMenuScaleSlider_OnValueChanged(self)
	if HatRackPerOptions then
		HatRackPerOptions.MenuScale = self:GetValue()
		HatRack_OptMenuScaleSliderText:SetText(format("Menu Scale: %.2f",HatRackPerOptions.MenuScale))
		HatRack_MenuFrame:SetScale(HatRackPerOptions.MenuScale)
	end
end

function HatRack.CheckButton_OnClick(self)
	local _,_,var = string.find(self:GetName(),"HatRack_Opt(.+)")
	if HatRackOptions[var] then
		HatRackOptions[var] = self:GetChecked() and "ON" or "OFF"
		PlaySound(self:GetChecked() and "igMainMenuOptionCheckBoxOn" or "igMainMenuOptionCheckBoxOff")
		HatRack.ValidateChecks()
	end

	if self==HatRack_OptCooldownCount then
		HatRack.WriteWornCooldowns()
		HatRack.WriteMenuCooldowns()
	elseif self==HatRack_OptLocked then
		HatRack.DockWindows()
		HatRack.ReflectLock()
	elseif self==HatRack_OptKeepOpen or self==HatRack_OptSetColumns then
		if HatRackOptions.KeepOpen=="ON" then
			HatRack.BuildMenu()
		end
	elseif self==HatRack_OptKeepDocked then
		HatRack.DockWindows()
	elseif self==HatRack_OptLargeCooldown then
		HatRack.ReflectCooldownFont()
	elseif self==HatRack_OptSquareMinimap then
		HatRack.MoveMinimapButton()
	elseif self==HatRack_OptShowHotKeys then
		HatRack.KeyBindingsChanged()
	elseif self==HatRack_OptShowIcon then
		HatRack.MoveMinimapButton()
	elseif self==HatRack_OptRedRange then
		HatRack.ReflectRedRange()
	elseif self==HatRack_OptMenuOnRight then
		HatRack.ReflectMenuOnRight()
	end
end

function HatRack.ReflectLock()
	local c = HatRackOptions.Locked=="ON" and 0 or .5
	HatRack_OptFrame:SetBackdropBorderColor(c,c,c,1)
	HatRack_MainFrame:SetBackdropColor(c,c,c,c)
	HatRack_MainFrame:SetBackdropBorderColor(c,c,c,c*2)
	HatRack_MenuFrame:SetBackdropColor(c,c,c,c)
	HatRack_MenuFrame:SetBackdropBorderColor(c,c,c,c*2)
	HatRack_MenuFrame:EnableMouse(c*2)
	HatRack_OptLocked:SetChecked(1-c*2)
	local normalTexture = HatRack_LockButton:GetNormalTexture()
	local pushedTexture = HatRack_LockButton:GetPushedTexture()
	local mouseDown, mouseUp
	if c==0 then
		normalTexture:SetTexCoord(.875,1,.125,.25)
		pushedTexture:SetTexCoord(.75,.875,.125,.25)
	else
		mouseDown = function() this:StartMoving() end
		mouseUp = function() this:StopMovingOrSizing() end
		normalTexture:SetTexCoord(.75,.875,.125,.25)
		pushedTexture:SetTexCoord(.875,1,.125,.25)
	end
	HatRack_OptFrame:SetScript("OnMouseDown", mouseDown)
	HatRack_OptFrame:SetScript("OnMouseUp", mouseUp)
end

function HatRack.ReflectCooldownFont()
	HatRack.SetCooldownFont("HatRack_Hat")
	for i=1,30 do
		HatRack.SetCooldownFont("HatRack_Menu"..i)
	end
end

function HatRack.SetCooldownFont(button)
	local item = getglobal(button.."Time")
	if HatRackOptions.LargeCooldown=="ON" then
		item:SetFont("Fonts\\FRIZQT__.TTF",16,"OUTLINE")
		item:SetTextColor(1,.82,0,1)
		item:ClearAllPoints()
		item:SetPoint("CENTER",button,"CENTER")
	else
		item:SetFont("Fonts\\ARIALN.TTF",14,"OUTLINE")
		item:SetTextColor(1,1,1,1)
		item:ClearAllPoints()
		item:SetPoint("BOTTOM",button,"BOTTOM")
	end
end


--[[ Titlebar buttons ]]

function HatRack.SmallButton_OnClick(self)
	PlaySound("igMainMenuOptionCheckBoxOn")
	if self==HatRack_CloseButton then
		HatRack_OptFrame:Hide()
	elseif self==HatRack_LockButton then
		HatRackOptions.Locked = (HatRackOptions.Locked=="ON") and "OFF" or "ON"
		HatRack.DockWindows()
		HatRack.ReflectLock()
	end
end